#!/usr/bin/perl -w

=begin comment info
 +-----------------------------------------------------------------------------+
 | Copyright 2000-2005 Guardian Digital, Inc., All Rights Reserved
 | $Id: named-upgrade-3.0.pl,v 1.5 2005/09/01 16:33:45 pete Exp $
 |
 | Peter J. O'Hara <pete@guardiandigital.com>
 | Pax Dickenson <pax@guardiandigital.com>
 +-----------------------------------------------------------------------------+
=end comment info
=cut

  # use perl;
  use WebTool;
  use POSIX qw(WEXITSTATUS);
  use strict;
  use warnings;
  
  use constant NAMED_CONF     => '/var/chroot/named/var/named/named.conf';

  #
  # Load our 'named' module before we go any further.
  #
  BEGIN { WebTool::load_module('named'); };

  #
  # Local prototypes.
  #
  sub get_zone_name($);
  sub get_zone_file($);
  sub retcode($);
  sub backup_file($$$);
  sub update_conf($);
  sub check_new_conf();

  #
  # XXX
  #
  my $named	= new named;
  my @zones	= get_zone_name($named);
  my $tmp	= {};
  my $changed	= undef;
  my @TMP	= ();
  my $search    = "# This was added for security purposes under BIND 9";

  # Check and see if named has already been upgraded
  open NAMED, NAMED_CONF or die "Can't open named.conf: $!";
  while (my $line = <NAMED>) {
    chomp($line);
    push @TMP, $line;
  }
  close NAMED;

  #
  # Exit quietly if named has already been upgraded.
  #
  foreach my $line (@TMP) { exit 0 if ($line =~ /^.*$search.*$/); } 

  # Update named.conf
  update_conf(NAMED_CONF);

  # Now update master zone files
#
# XXX - this assumes that only master zones are returned from zone_get()
#

#
# XXX - this script assumes that all zone files in named.conf exist
#

  foreach my $z (@zones) {
    my $zone_data = $named->zone_get($z);
    my $z_file = $zone_data->{'filename'};
    my $z_origin = $zone_data->{'soa'}->{'ORIGIN'};
    my @file = ();
    my $r = ref($zone_data->{'records'}->{'NS'});

    # Read in zone file
    open (FILE, $z_file) or die "Couldn't open file $z: $!";
    while ( my $line = <FILE> ) {
      chomp($line);
      push @file, $line; 
    }
    close FILE;

    foreach my $line (@file) {
      # Find SOA record
      if ($line eq "") {			# Remove emtpy lines
        next;
      } elsif ($line =~ /^(.*)(IN\s+SOA\s+.*$)/) {
        if (!defined($z_origin)) {		# Needs a $ORIGIN record
          push @{$tmp->{$z}}, "\$ORIGIN	$z.";
          push @{$tmp->{$z}}, "";
          $changed = 1;
        }
        push @{$tmp->{$z}}, "@	$2";
      } elsif (($line =~ /\s+MX\s+/ || $line =~ /\s+NS\s+/)
            && $line =~ /^(.*)(\s+IN\s+.*)$/) {
        push @{$tmp->{$z}}, "@	$2";
        $changed = 1;
      } elsif ($line =~ /^\$ORIGIN\s+.*$/) {	# If $ORIGIN follow with blank
          push @{$tmp->{$z}}, $line;
          push @{$tmp->{$z}}, '';
          $changed = 1;
      } elsif ($line =~ /^$z\.?\s+.*$/ && $line =~ /^(.*)(\s+IN\s+A\s+.*)$/) {
          push @{$tmp->{$z}}, "@	$2";	# If domain has an IP use "@"
      } elsif ($line =~ /^\s+(IN.*)$/) {	# If nothing before "IN" add @
          push @{$tmp->{$z}}, "@	$1";
          $changed = 1;
      } elsif ($line =~ /^(.*)\s+(IN\s+.*)$/) {
          my $rem = $2;

          if ($1 =~ /^([^\..]+)\..*$/) {		# Strip domain name off
            my $host = $1;
            push @{$tmp->{$z}}, "$host	$rem";
            $changed = 1;
          } else {
            push @{$tmp->{$z}}, $line;
            $changed = 1;
          }
        } else {
          push @{$tmp->{$z}}, $line;
        }
    }

    # If anything changed create a new zone file and replace old one
    if ($changed) {
      my $new_file = "$z_file.new";

      open (NEW_FILE, "> $new_file") or die "Couldn't write to $new_file: $!";
      foreach my $line (@{$tmp->{$z}}) {
        print NEW_FILE "$line\n";
      }
      close NEW_FILE;
 
      my $z_file_sav = $z_file;
      $z_file_sav .= ".sav";
      my $rv = system("mv", "-f", $z_file, $z_file_sav);
      retcode($rv);
      $rv = system("mv", "-f", $new_file, $z_file);
      retcode($rv);

      # If we got here I assume zone files were all made - can remove the saved
      # copies
      system("rm", $z_file_sav);
    }
  }
  
################################################################################
sub retcode($) {
  my $rv = shift;
  if (WEXITSTATUS($rv) != 0) {
    print "Couldn't mv file!\n";
    exit 1;
  }
}

################################################################################
sub get_zone_name($) {
  my $named	= shift;
  my $cfg	= $named->read_named_conf();
  my @rv	= ();

  foreach my $z (sort keys %{$cfg->{'zones'}}) {
    my $zone	= $cfg->{'zones'}->{$z};

    next unless ($zone->{'type'} eq 'master');

    push @rv, $zone->{'name'};
  }

  return @rv;
}

################################################################################
sub get_zone_file($) {
  my $named	= shift;
  my $cfg	= $named->read_named_conf();
  my @rv	= ();

  foreach my $z (sort keys %{$cfg->{'zones'}}) {
    my $zone	= $cfg->{'zones'}->{$z};

    next unless ($zone->{'type'} eq 'master');

    push @rv, named::NAMED_CONFDIR . '/' . $zone->{'file'};
  }

  return @rv;
}

################################################################################
sub backup_file($$$) {
  my $filename = shift;
  my $bak = shift;
  my $count = shift;

  if (! -e $filename.$bak ) {
    system("cp", "-p", $filename, $filename.$bak);
  }
  elsif (! -e $filename.$bak.$count ) {
    system("cp", "-p", $filename, $filename.$bak.$count);
  }
  else {
    backup_file($filename, $bak, ++$count)
  }
}

################################################################################
sub update_conf($) {
# this should be run prior to upgrading bind.
  my $filename = shift;

  my $bak = ".gdbackup";
  my @f;
  my $badline;
  my $warning;
  my $zonechaos = 0;
  my $sectoptions = 0;


  # Read in file
  open (CONF, $filename) || die "Could not read config file ".$filename;
  while (my $cl = <CONF>) { chomp $cl; push @f, $cl; }
  close CONF;

  print STDERR "Checking $filename\n";

  # the main loop
  for (my $i=0; $i < @f; $i++) {
    if ($f[$i] =~ m/^\s*(#|\/\/)/i) {
        # skip comments, indicated by leading # or //
    }
    elsif ($f[$i] =~ m/(\s*)(zone\s+"bind"\s+chaos\s+{)(.*)/i) {
      # this zone is used in BIND 8 only
      $warning = "The chaos zone is invalid under BIND 9.";
      $badline = $f[$i];
      print STDERR "Line $i: $warning\n";
      splice (@f, $i++, 0, "# $warning");
      $zonechaos = 1;
    }    
    elsif ($f[$i] =~ m/(\s*)(options\s+{)(.*)/i) {
      # finding options block for addition
      $sectoptions = $i+1;
    }
    elsif ($f[$i] =~ m/^(\s*)(}\s*;)\s*$/i) {
      # end of block
      if ($zonechaos) {
        $zonechaos = 0;
        $f[$i] = "# ".$f[$i];
      } 
    }
    else { 
      # OK line, skip it
    }

    if ($zonechaos) {
      # within the chaos zone, need to comment out
      $f[$i] = "# ".$f[$i];
    }
  }
  
  # Need to add version "None of your business."
  splice (@f, $sectoptions++, 0, "\t# This was added for security purposes under BIND 9.");
  splice (@f, $sectoptions++, 0, "\tversion \"None of your business.\";");
  splice (@f, $sectoptions++, 0, "");

  # backup original file with cp -p
  backup_file($filename, $bak, 0);

  # Write out file
  open (CONF, '>', $filename) || die "Could not write config file $filename";
  foreach my $l (@f) { print CONF "$l\n"; }
  close CONF;

}

################################################################################
sub check_new_conf() {
  # this function should only be run after updating to BIND 9
  # it will check the configuration so the user can be alerted to any problem

  # not sure how to run this, if we want to send output to the user?
  # will wait until we can test the upgrade
  # this function will be very simple anyway

}
